# This file is part of Checkbox.
#
# Copyright 2016-2019 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
Bunch of helpers to make it easier to reuse logic and UI functionality
associated with a particular stage of checkbox execution.
"""
import abc
import datetime
import gettext
import json
import logging
import os
import time

from plainbox.abc import IJobResult
from plainbox.i18n import pgettext as C_
from plainbox.impl.result import JobResultBuilder
from plainbox.impl.result import tr_outcome
from plainbox.impl.transport import InvalidSecureIDError
from plainbox.impl.transport import TransportError
from plainbox.impl.transport import get_all_transports
from plainbox.impl.unit.exporter import ExporterError

from checkbox_ng.launcher.run import (
    Action, ActionUI, NormalUI, ReRunJob, seconds_to_human_duration)

_ = gettext.gettext

_logger = logging.getLogger("checkbox-ng.launcher.stages")


class CheckboxUiStage(metaclass=abc.ABCMeta):
    """
    Functionality governing command job execution.
    """
    @property
    @abc.abstractmethod
    def sa(self):
        """SessionAssistant instance to use."""

    @property
    @abc.abstractmethod
    def C(self):
        """Colorizer instance to use."""

    @property
    @abc.abstractmethod
    def is_interactive(self):
        """
        Flag indicating that this is an interactive invocation.

        We can then interact with the user when we encounter OUTCOME_UNDECIDED.
        """

    def _pick_action_cmd(self, action_list, prompt=None):
        return ActionUI(action_list, prompt).run()


class MainLoopStage(CheckboxUiStage):

    def __init__(self):
        super().__init__()
        self._sudo_password = None
        self._passwordless_sudo = False
        self._reset_auto_submission_retries()

    def _reset_auto_submission_retries(self):
        """
        This is outlined so both, __init__ and later part of the logic
        can set the same value.
        """
        self._auto_submission_retries = 3


    def _run_single_job_with_ui_loop(self, job, ui):
        print(self.C.header(job.tr_summary(), fill='-'))
        print(_("ID: {0}").format(job.id))
        print(_("Category: {0}").format(
            self.sa.get_job_state(job.id).effective_category_id))
        comments = ""
        while True:
            if job.plugin in ('user-interact', 'user-interact-verify',
                              'user-verify', 'manual'):
                job_state = self.sa.get_job_state(job.id)
                if (not self.is_interactive and
                        job.plugin in ('user-interact',
                                       'user-interact-verify',
                                       'manual')):
                    result_builder = JobResultBuilder(
                        outcome=IJobResult.OUTCOME_SKIP,
                        comments=_("Trying to run interactive job in a silent"
                                   " session"))
                    return result_builder
                if job_state.can_start():
                    ui.notify_about_purpose(job)
                if (self.is_interactive and
                        job.plugin in ('user-interact',
                                       'user-interact-verify',
                                       'manual')):
                    if job_state.can_start():
                        ui.notify_about_steps(job)
                    if job.plugin == 'manual':
                        cmd = 'run'
                    else:
                        if job_state.can_start():
                            cmd = ui.wait_for_interaction_prompt(job)
                        else:
                            # 'running' the job will make it marked as skipped
                            # because of the failed dependency
                            cmd = 'run'
                    if cmd == 'run' or cmd is None:
                        result_builder = self.sa.run_job(job.id, ui, False)
                    elif cmd == 'comment':
                        new_comment = input(self.C.BLUE(
                            _('Please enter your comments:') + '\n'))
                        if new_comment:
                            comments += new_comment + '\n'
                        continue
                    elif cmd == 'skip':
                        result_builder = JobResultBuilder(
                            outcome=IJobResult.OUTCOME_SKIP,
                            comments=_("Explicitly skipped before"
                                       " execution"))
                        if comments != "":
                            result_builder.comments = comments
                        break
                    elif cmd == 'quit':
                        raise SystemExit()
                else:
                    result_builder = self.sa.run_job(job.id, ui, False)
            else:
                if 'noreturn' in job.get_flag_set():
                    ui.noreturn_job()
                result_builder = self.sa.run_job(job.id, ui, False)
            if (self.is_interactive and
                    result_builder.outcome == IJobResult.OUTCOME_UNDECIDED):
                try:
                    if comments != "":
                        result_builder.comments = comments
                    ui.notify_about_verification(job)
                    self._interaction_callback(job, result_builder)
                except ReRunJob:
                    self.sa.use_job_result(job.id, result_builder.get_result())
                    continue
            break
        return result_builder

    def _interaction_callback(self, job, result_builder,
                              prompt=None, allowed_outcome=None):
        result = result_builder.get_result()
        if prompt is None:
            prompt = _("Select an outcome or an action: ")
        if allowed_outcome is None:
            allowed_outcome = [IJobResult.OUTCOME_PASS,
                               IJobResult.OUTCOME_FAIL,
                               IJobResult.OUTCOME_SKIP]
        allowed_actions = [
            Action('c', _('add a comment'), 'set-comments')
        ]
        if IJobResult.OUTCOME_PASS in allowed_outcome:
            allowed_actions.append(
                Action('p', _('set outcome to {0}').format(
                    self.C.GREEN(C_('set outcome to <pass>', 'pass'))),
                    'set-pass'))
        if IJobResult.OUTCOME_FAIL in allowed_outcome:
            allowed_actions.append(
                Action('f', _('set outcome to {0}').format(
                    self.C.RED(C_('set outcome to <fail>', 'fail'))),
                    'set-fail'))
        if IJobResult.OUTCOME_SKIP in allowed_outcome:
            allowed_actions.append(
                Action('s', _('set outcome to {0}').format(
                    self.C.YELLOW(C_('set outcome to <skip>', 'skip'))),
                    'set-skip'))
        if job.command is not None:
            allowed_actions.append(
                Action('r', _('re-run this job'), 're-run'))
        if result.return_code is not None:
            if result.return_code == 0:
                suggested_outcome = IJobResult.OUTCOME_PASS
            else:
                suggested_outcome = IJobResult.OUTCOME_FAIL
            allowed_actions.append(
                Action('', _('set suggested outcome [{0}]').format(
                    tr_outcome(suggested_outcome)), 'set-suggested'))
        while result.outcome not in allowed_outcome:
            print(_("Please decide what to do next:"))
            print("  " + _("outcome") + ": {0}".format(
                self.C.result(result)))
            if result.comments is None:
                print("  " + _("comments") + ": {0}".format(
                    C_("none comment", "none")))
            else:
                print("  " + _("comments") + ": {0}".format(
                    self.C.CYAN(result.comments, bright=False)))
            cmd = self._pick_action_cmd(allowed_actions)
            if cmd == 'set-pass':
                result_builder.outcome = IJobResult.OUTCOME_PASS
            elif cmd == 'set-fail':
                result_builder.outcome = IJobResult.OUTCOME_FAIL
            elif cmd == 'set-skip' or cmd is None:
                result_builder.outcome = IJobResult.OUTCOME_SKIP
            elif cmd == 'set-suggested':
                result_builder.outcome = suggested_outcome
            elif cmd == 'set-comments':
                new_comment = input(self.C.BLUE(
                    _('Please enter your comments:') + '\n'))
                if new_comment:
                    result_builder.add_comment(new_comment)
            elif cmd == 're-run':
                raise ReRunJob
            result = result_builder.get_result()

    def _get_ui_for_job(self, job):
        class CheckboxUI(NormalUI):
            def considering_job(self, job, job_state):
                pass
        show_out = True
        if 'suppress-output' in job.get_flag_set():
            show_out = False
        return CheckboxUI(self.C.c, show_cmd_output=show_out)

    def _run_jobs(self, jobs_to_run):
        estimated_time = 0
        for job_id in jobs_to_run:
            job = self.sa.get_job(job_id)
            if (job.estimated_duration is not None and
                    estimated_time is not None):
                estimated_time += job.estimated_duration
            else:
                estimated_time = None
        for job_no, job_id in enumerate(jobs_to_run, start=1):
            print(self.C.header(
                _('Running job {} / {}. Estimated time left: {}').format(
                    job_no, len(jobs_to_run),
                    seconds_to_human_duration(max(0, estimated_time))
                    if estimated_time is not None else _("unknown")),
                fill='-'))
            job = self.sa.get_job(job_id)
            builder = self._run_single_job_with_ui_loop(
                job, self._get_ui_for_job(job))
            result = builder.get_result()
            self.sa.use_job_result(job_id, result)
            if (job.estimated_duration is not None and
                    estimated_time is not None):
                estimated_time -= job.estimated_duration

    def _run_bootstrap_jobs(self, jobs_to_run):
        for job_no, job_id in enumerate(jobs_to_run, start=1):
            print(self.C.header(
                _('Bootstrap {} ({}/{})').format(
                    job_id, job_no, len(jobs_to_run), fill='-')))
            result_builder = self.sa.run_job(job_id, 'piano', False)
            self.sa.use_job_result(job_id, result_builder.get_result())

    def _generate_job_infos(self, job_list):
        test_info_list = tuple()
        for job in job_list:
            cat_id = self.sa.get_job_state(job.id).effective_category_id
            duration_txt = _('No estimated duration provided for this job')
            if job.estimated_duration is not None:
                duration_txt = '{} {}'.format(job.estimated_duration, _(
                    'seconds'))
            test_info = {
                "id": job.id,
                "partial_id": job.partial_id,
                "name": job.tr_summary(),
                "category_id": cat_id,
                "category_name": self.sa.get_category(cat_id).tr_name(),
                "automated": (
                    _('this job is fully automated')
                    if job.automated
                    else _('this job requires some manual interaction')
                ),
                "duration": duration_txt,
                "description": (job.tr_description() or
                                _('No description provided for this job')),
                "outcome": self.sa.get_job_state(job.id).result.outcome,
            }
            test_info_list = test_info_list + ((test_info, ))
        return test_info_list

    def _generate_tp_infos(self, tp_list):
        tp_info_list = []
        for tp_id in tp_list:
            tp_info = {
                'id': tp_id,
                'name': self.sa.get_test_plan(tp_id).name
            }
            tp_info_list.append(tp_info)
        return tp_info_list


class ReportsStage(CheckboxUiStage):

    def __init__(self):
        super().__init__()
        self._export_fn = None

    def _override_exporting(self, export_fn):
        self._export_fn = export_fn

    def _prepare_stock_report(self, report):
        # this is purposefully not using pythonic dict-keying for better
        # readability
        if not self.sa.config.transports:
            self.sa.config.transports = dict()
        if not self.sa.config.exporters:
            self.sa.config.exporters = dict()
        if not self.sa.config.reports:
            self.sa.config.reports = dict()
        if report == 'text':
            self.sa.config.exporters['text'] = {
                'unit': 'com.canonical.plainbox::text'}
            self.sa.config.transports['stdout'] = {
                'type': 'stream', 'stream': 'stdout'}
            # '1_' prefix ensures ordering amongst other stock reports. This
            # report name does not appear anywhere (because of forced: yes)
            self.sa.config.reports['1_text_to_screen'] = {
                'transport': 'stdout', 'exporter': 'text', 'forced': 'yes'}
        elif report == 'certification':
            self.sa.config.exporters['tar'] = {
                'unit': 'com.canonical.plainbox::tar'}
            self.sa.config.transports['c3'] = {
                'type': 'submission-service',
                'secure_id': self.sa.config.transports.get('c3', {}).get(
                    'secure_id', None)}
            self.sa.config.reports['upload to certification'] = {
                'transport': 'c3', 'exporter': 'tar'}
        elif report == 'certification-staging':
            self.sa.config.exporters['tar'] = {
                'unit': 'com.canonical.plainbox::tar'}
            self.sa.config.transports['c3-staging'] = {
                'type': 'submission-service',
                'secure_id': self.sa.config.transports.get('c3', {}).get(
                    'secure_id', None),
                'staging': 'yes'}
            self.sa.config.reports['upload to certification-staging'] = {
                'transport': 'c3-staging', 'exporter': 'tar'}
        elif report == 'submission_files':
            # LP:1585326 maintain isoformat but removing ':' chars that cause
            # issues when copying files.
            isoformat = "%Y-%m-%dT%H.%M.%S.%f"
            timestamp = datetime.datetime.utcnow().strftime(isoformat)
            if not os.path.exists(self.base_dir):
                os.makedirs(self.base_dir)
            for exporter, file_ext in [('html', '.html'),
                                       ('junit', '.junit.xml'),
                                       ('tar', '.tar.xz')]:
                path = os.path.join(self.base_dir, ''.join(
                    ['submission_', timestamp, file_ext]))
                self.sa.config.transports['{}_file'.format(exporter)] = {
                    'type': 'file',
                    'path': path}
                if exporter not in self.sa.config.exporters:
                    self.sa.config.exporters[exporter] = {
                        'unit': 'com.canonical.plainbox::{}'.format(
                            exporter)}
                if not self.sa.config.exporters[exporter].get('unit'):
                    unit = 'com.canonical.plainbox::{}'.format(exporter)
                    self.sa.config.exporters[exporter]['unit'] = unit
                self.sa.config.reports['2_{}_file'.format(exporter)] = {
                    'transport': '{}_file'.format(exporter),
                    'exporter': '{}'.format(exporter),
                    'forced': 'yes'
                }

    def _prepare_transports(self):
        self.base_dir = os.path.join(
            os.getenv(
                'XDG_DATA_HOME', os.path.expanduser("~/.local/share/")),
            "checkbox-ng")
        self._available_transports = get_all_transports()
        self.transports = dict()

    def _create_transport(self, transport):
        if transport in self.transports:
            return
        # depending on the type of transport we need to pick variable that
        # serves as the 'where' param for the transport. In case of
        # certification site the URL is supplied here
        tr_type = self.sa.config.transports[transport]['type']
        if tr_type not in self._available_transports:
            _logger.error(_("Unrecognized type '%s' of transport '%s'"),
                          tr_type, transport)
            raise SystemExit(1)
        cls = self._available_transports[tr_type]
        if tr_type == 'file':
            self.transports[transport] = cls(
                os.path.expanduser(
                    self.sa.config.transports[transport]['path']))
        elif tr_type == 'stream':
            self.transports[transport] = cls(
                self.sa.config.transports[transport]['stream'])
        elif tr_type == 'submission-service':
            secure_id = self.sa.config.transports[transport].get(
                'secure_id', None)
            if self.is_interactive:
                new_description = input(self.C.BLUE(_(
                    'Enter submission description (press Enter to skip): ')))
                if new_description:
                    self.sa.update_app_blob(json.dumps(
                        {
                            'description': new_description,

                        }).encode("UTF-8"))
            if not secure_id and self.is_interactive:
                secure_id = input(self.C.BLUE(_('Enter secure-id:')))
            if secure_id:
                options = "secure_id={}".format(secure_id)
            else:
                options = ""
            if self.sa.config.transports[transport].get('staging', False):
                url = ('https://certification.staging.canonical.com/'
                       'api/v1/submission/{}/'.format(secure_id))
            elif os.getenv('C3_URL'):
                url = (
                    '{}/{}/'.format(os.getenv('C3_URL'), ctx.args.secure_id))
            else:
                url = ('https://certification.canonical.com/'
                       'api/v1/submission/{}/'.format(secure_id))
            self.transports[transport] = cls(url, options)

    def _export_results(self):
        if 'none' not in self.sa.config.stock_reports:
            for report in self.sa.config.stock_reports:
                if report in ['certification', 'certification-staging']:
                    # skip stock c3 report if secure_id is not given from
                    # config files or launchers, and the UI is non-interactive
                    # (silent)
                    if ('c3' not in self.sa.config.transports and
                            not self.is_interactive):
                        continue
                    # don't generate stock c3 reports if sideloaded providers
                    # were in use, something that should only be done during
                    # development
                    if self.sa.sideloaded_providers:
                        _logger.warning(_("Using side-loaded providers "
                                          "disabled the %s report"), report)
                        continue
                self._prepare_stock_report(report)
        # reports are stored in an ordinary dict(), so sorting them ensures
        # the same order of submitting them between runs, and if they
        # share common prefix, they are next to each other
        for name, params in sorted(self.sa.config.reports.items()):
            # don't generate stock c3 reports if sideloaded providers
            # were in use, something that should only be done during
            # development
            if (params.get('transport') == 'certification' and
                    self.sa.sideloaded_providers):
                _logger.warning(_("Using side-loaded providers disabled "
                                  "the %s report"), name)
                continue
            if self.is_interactive and not params.get('forced', False):
                message = _("Do you want to submit '{}' report?").format(name)
                cmd = self._pick_action_cmd([
                    Action('y', _("yes"), 'y'),
                    Action('n', _("no"), 'n')
                ], message)
            else:
                cmd = 'y'
            if cmd == 'n':
                continue
            exporter_id = self.sa.config.exporters[params['exporter']]['unit']
            exporter_options = self.sa.config.exporters[
                    params['exporter']].get('options', '').split()
            done_sending = False
            while not done_sending:
                try:
                    self._create_transport(params['transport'])
                    transport = self.transports[params['transport']]
                    if self._export_fn:
                        result = self._export_fn(exporter_id, transport)
                    else:
                        try:
                            result = self.sa.export_to_transport(
                                exporter_id, transport, exporter_options)
                        except ExporterError as exc:
                            _logger.warning(
                                _("Problem occured when preparing %s report:"
                                  "%s"), exporter_id, exc)
                    if result and 'url' in result:
                        print(result['url'])
                    elif result and 'status_url' in result:
                        print(result['status_url'])
                except TransportError as exc:
                    _logger.warning(
                        _("Problem occured when submitting '%s' report: %s"),
                        name, exc)
                    if self._retry_dialog():
                        # let's remove current transport, so in next
                        # iteration it will be "rebuilt", so if some parts
                        # were user-provided, checkbox will ask for them
                        # again
                        self.transports.pop(params['transport'])
                        continue
                except InvalidSecureIDError:
                    _logger.warning(_("Invalid secure_id"))
                    if not self.is_interactive:
                        # secure_id will not magically change if the session
                        # is a non-interactive one, so let's stop trying
                        done_sending = True
                        continue
                    if self._retry_dialog():
                        self.sa.config.transports['c3'].pop('secure_id')
                        continue
                except Exception:
                    _logger.error(
                        _("Problem with a '%s' report using '%s' exporter "
                          "sent to '%s' transport."),
                        name, exporter_id, transport.url)
                self._reset_auto_submission_retries()
                done_sending = True

    def _retry_dialog(self):
        if self.is_interactive:
            message = _("Do you want to retry?")
            cmd = self._pick_action_cmd([
                Action('y', _("yes"), 'y'),
                Action('n', _("no"), 'n')
            ], message)
            if cmd == 'y':
                return True
        else:
            if self._auto_submission_retries:
                self._auto_submission_retries -= 1
                # let's double the sleep length with each retry
                sleep_length = [120, 60, 30][self._auto_submission_retries]
                print("Retrying in {}s".format(sleep_length))
                time.sleep(sleep_length)
                return True

        return False



# This file is part of Checkbox.
#
# Copyright 2016-2019 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
Definition of sub-command classes for checkbox-cli
"""
from argparse import ArgumentTypeError
from argparse import SUPPRESS
from collections import defaultdict
from string import Formatter
from tempfile import TemporaryDirectory
import copy
import fnmatch
import gettext
import json
import logging
import operator
import os
import re
import sys
import tarfile
import time

from plainbox.abc import IJobResult
from plainbox.i18n import ngettext
from plainbox.impl.color import Colorizer
from plainbox.impl.execution import UnifiedRunner
from plainbox.impl.highlevel import Explorer
from plainbox.impl.result import MemoryJobResult
from plainbox.impl.runner import slugify
from plainbox.impl.secure.config import Unset
from plainbox.impl.secure.sudo_broker import sudo_password_provider
from plainbox.impl.session.assistant import SA_RESTARTABLE
from plainbox.impl.session.restart import detect_restart_strategy
from plainbox.impl.session.restart import get_strategy_by_name
from plainbox.impl.session.storage import WellKnownDirsHelper
from plainbox.impl.transport import TransportError
from plainbox.impl.transport import get_all_transports
from plainbox.impl.transport import SECURE_ID_PATTERN

from checkbox_ng.config import load_configs
from checkbox_ng.launcher.stages import MainLoopStage, ReportsStage
from checkbox_ng.launcher.startprovider import (
    EmptyProviderSkeleton, IQN, ProviderSkeleton)
from checkbox_ng.launcher.run import Action
from checkbox_ng.launcher.run import NormalUI
from checkbox_ng.urwid_ui import CategoryBrowser
from checkbox_ng.urwid_ui import ManifestBrowser
from checkbox_ng.urwid_ui import ReRunBrowser
from checkbox_ng.urwid_ui import TestPlanBrowser

_ = gettext.gettext

_logger = logging.getLogger("checkbox-ng.launcher.subcommands")


class Submit():
    def register_arguments(self, parser):
        def secureid(secure_id):
            if not re.match(SECURE_ID_PATTERN, secure_id):
                raise ArgumentTypeError(
                    _("must be 15-character (or more) alphanumeric string"))
            return secure_id
        parser.add_argument(
            'secure_id', metavar=_("SECURE-ID"),
            type=secureid,
            help=_("associate submission with a machine using this SECURE-ID"))
        parser.add_argument(
            "submission", metavar=_("SUBMISSION"),
            help=_("The path to the results file"))
        parser.add_argument(
            "-s", "--staging", action="store_true",
            help=_("Use staging environment"))
        parser.add_argument(
            "-m", "--message",
            help=_("Submission description"))

    def invoked(self, ctx):
        transport_cls = None
        mode = 'rb'
        options_string = "secure_id={0}".format(ctx.args.secure_id)
        url = ('https://certification.canonical.com/'
               'api/v1/submission/{}/'.format(ctx.args.secure_id))
        submission_file = ctx.args.submission
        if ctx.args.staging:
            url = ('https://certification.staging.canonical.com/'
                   'api/v1/submission/{}/'.format(ctx.args.secure_id))
        elif os.getenv('C3_URL'):
            url = ('{}/{}/'.format(os.getenv('C3_URL'), ctx.args.secure_id))
        from checkbox_ng.certification import SubmissionServiceTransport
        transport_cls = SubmissionServiceTransport
        transport = transport_cls(url, options_string)
        if ctx.args.message:
            tmpdir = TemporaryDirectory()
            with tarfile.open(ctx.args.submission) as tar:
                tar.extractall(tmpdir.name)
            with open(os.path.join(tmpdir.name, 'submission.json')) as f:
                json_payload = json.load(f)
            with open(os.path.join(tmpdir.name, 'submission.json'), 'w') as f:
                json_payload['description'] = ctx.args.message
                json.dump(json_payload, f, sort_keys=True, indent=4)
            new_subm_file = os.path.join(
                tmpdir.name, os.path.basename(ctx.args.submission))
            with tarfile.open(new_subm_file, mode='w:xz') as tar:
                tar.add(tmpdir.name, arcname='')
            submission_file = new_subm_file
        try:
            with open(submission_file, mode) as subm_file:
                result = transport.send(subm_file)
        except (TransportError, OSError) as exc:
            raise SystemExit(exc)
        else:
            if result and 'url' in result:
                # TRANSLATORS: Do not translate the {} format marker.
                print(_("Successfully sent, submission status"
                        " at {0}").format(result['url']))
            elif result and 'status_url' in result:
                # TRANSLATORS: Do not translate the {} format marker.
                print(_("Successfully sent, submission status"
                        " at {0}").format(result['status_url']))
            else:
                # TRANSLATORS: Do not translate the {} format marker.
                print(_("Successfully sent, server response"
                        ": {0}").format(result))


class StartProvider():
    def register_arguments(self, parser):
        parser.add_argument(
            'name', metavar=_('name'), type=IQN,
            # TRANSLATORS: please keep the YYYY.example... text unchanged or at
            # the very least translate only YYYY and some-name. In either case
            # some-name must be a reasonably-ASCII string (should be safe for a
            # portable directory name)
            help=_("provider name, eg: YYYY.example.org:some-name"))
        parser.add_argument(
            '--empty', action='store_const', const=EmptyProviderSkeleton,
            default=ProviderSkeleton, dest='skeleton',
            help=_('create an empty provider'))

    def invoked(self, ctx):
        ctx.args.skeleton(ctx.args.name).instantiate(
            '.', name=ctx.args.name,
            gettext_domain=re.sub("[.:]", "_", ctx.args.name))


class Launcher(MainLoopStage, ReportsStage):
    @property
    def sa(self):
        return self.ctx.sa

    @property
    def C(self):
        return self._C

    def get_sa_api_version(self):
        return self.launcher.api_version

    def get_sa_api_flags(self):
        return self.launcher.api_flags

    def invoked(self, ctx):
        if ctx.args.version:
            from checkbox_ng.version import get_version_info
            for component, version in get_version_info().items():
                print("{}: {}".format(component, version))
            return
        if ctx.args.verify:
            # validation is always run, so if there were any errors the program
            # exited by now, so validation passed
            print(_("Launcher seems valid."))
            return
        self.launcher = load_configs(ctx.args.launcher)
        logging_level = {
            'normal': logging.WARNING,
            'verbose': logging.INFO,
            'debug': logging.DEBUG,
        }[self.launcher.verbosity]
        if not ctx.args.verbose and not ctx.args.debug:
            # Command line args take precendence
            logging.basicConfig(level=logging_level)
        try:
            self._C = Colorizer()
            self.ctx = ctx
            # now we have all the correct flags and options, so we need to
            # replace the previously built SA with the defaults
            self._configure_restart(ctx)
            self._prepare_transports()
            ctx.sa.use_alternate_configuration(self.launcher)
            if not self._maybe_resume_session():
                self._start_new_session()
                self._pick_jobs_to_run()
            if not self.ctx.sa.get_static_todo_list():
                return 0
            if 'submission_files' in self.launcher.stock_reports:
                print("Reports will be saved to: {}".format(self.base_dir))
            # we initialize the nb of attempts for all the selected jobs...
            for job_id in self.ctx.sa.get_dynamic_todo_list():
                job_state = self.ctx.sa.get_job_state(job_id)
                job_state.attempts = self.launcher.max_attempts
            # ... before running them
            self._run_jobs(self.ctx.sa.get_dynamic_todo_list())
            if self.is_interactive and not self.launcher.auto_retry:
                while True:
                    if not self._maybe_rerun_jobs():
                        break
            elif self.launcher.auto_retry:
                while True:
                    if not self._maybe_auto_rerun_jobs():
                        break
            self._export_results()
            ctx.sa.finalize_session()
            return 0 if ctx.sa.get_summary()['fail'] == 0 else 1
        except KeyboardInterrupt:
            return 1

    @property
    def is_interactive(self):
        """
        Flag indicating that this is an interactive invocation.

        We can then interact with the user when we encounter OUTCOME_UNDECIDED.
        """
        return (self.launcher.ui_type == 'interactive' and
                sys.stdin.isatty() and sys.stdout.isatty())

    def _configure_restart(self, ctx):
        if SA_RESTARTABLE not in self.get_sa_api_flags():
            return
        if self.launcher.restart_strategy:
            try:
                cls = get_strategy_by_name(
                    self.launcher.restart_strategy)
                kwargs = copy.deepcopy(self.launcher.restart)
                # [restart] section has the kwargs for the strategy initializer
                # and the 'strategy' which is not one, let's pop it
                kwargs.pop('strategy')
                strategy = cls(**kwargs)
                ctx.sa.use_alternate_restart_strategy(strategy)

            except KeyError:
                _logger.warning(_('Unknown restart strategy: %s', (
                    self.launcher.restart_strategy)))
                _logger.warning(_(
                    'Using automatically detected restart strategy'))
                try:
                    strategy = detect_restart_strategy(session_type='local')
                except LookupError as exc:
                    _logger.warning(exc)
                    _logger.warning(_('Automatic restart disabled!'))
                    strategy = None
        else:
            strategy = detect_restart_strategy(session_type='local')
        if strategy:
            # gluing the command with pluses b/c the middle part
            # (launcher path) is optional
            snap_name = os.getenv('SNAP_NAME')
            if snap_name:
                # NOTE: This implies that any snap wishing to include a
                # Checkbox snap to be autostarted creates a snapcraft
                # app called "checkbox-cli"
                respawn_cmd = '/snap/bin/{}.checkbox-cli'.format(snap_name)
            else:
                respawn_cmd = sys.argv[0]  # entry-point to checkbox
            respawn_cmd += " launcher "
            if ctx.args.launcher:
                respawn_cmd += os.path.abspath(ctx.args.launcher) + ' '
            respawn_cmd += '--resume {}'  # interpolate with session_id
            ctx.sa.configure_application_restart(
                lambda session_id: [respawn_cmd.format(session_id)])

    def _maybe_resume_session(self):
        resume_candidates = list(self.ctx.sa.get_resumable_sessions())
        if self.ctx.args.session_id:
            requested_sessions = [s for s in resume_candidates if (
                s.id == self.ctx.args.session_id)]
            if requested_sessions:
                # session_ids are unique, so there should be only 1
                self._resume_session(requested_sessions[0])
                return True
            else:
                raise RuntimeError("Requested session is not resumable!")
        elif self.is_interactive:
            print(self.C.header(_("Resume Incomplete Session")))
            print(ngettext(
                "There is {0} incomplete session that might be resumed",
                "There are {0} incomplete sessions that might be resumed",
                len(resume_candidates)
            ).format(len(resume_candidates)))
            return self._run_resume_ui_loop(resume_candidates)
        else:
            return False

    def _run_resume_ui_loop(self, resume_candidates):
        for candidate in resume_candidates:
            cmd = self._pick_action_cmd([
                Action('r', _("resume this session"), 'resume'),
                Action('n', _("next session"), 'next'),
                Action('c', _("create new session"), 'create'),
                Action('d', _("delete old sessions"), 'delete'),
            ], _("Do you want to resume session {0!a}?").format(candidate.id))
            if cmd == 'next':
                continue
            elif cmd == 'create' or cmd is None:
                return False
            elif cmd == 'resume':
                self._resume_session(candidate)
                return True
            elif cmd == 'delete':
                ids = [candidate.id for candidate in resume_candidates]
                self._delete_old_sessions(ids)
                return False

    def _resume_session(self, session):
        metadata = self.ctx.sa.resume_session(session.id)
        if 'testplanless' not in metadata.flags:
            app_blob = json.loads(metadata.app_blob.decode("UTF-8"))
            test_plan_id = app_blob['testplan_id']
            self.ctx.sa.select_test_plan(test_plan_id)
            self.ctx.sa.bootstrap()
        last_job = metadata.running_job_name
        # If we resumed maybe not rerun the same, probably broken job
        self._handle_last_job_after_resume(last_job)

    def _start_new_session(self):
        print(_("Preparing..."))
        title = self.ctx.args.title or self.launcher.session_title
        title = title or self.launcher.app_id
        if self.launcher.app_version:
            title += ' {}'.format(self.launcher.app_version)
        runner_kwargs = {
            'normal_user_provider': lambda: self.launcher.normal_user,
            'password_provider': sudo_password_provider.get_sudo_password,
            'stdin': None,
        }
        self.ctx.sa.start_new_session(title, UnifiedRunner, runner_kwargs)
        if self.launcher.test_plan_forced:
            tp_id = self.launcher.test_plan_default_selection
            if tp_id not in self.ctx.sa.get_test_plans():
                _logger.error(_(
                    'The test plan "%s" is not available!'), tp_id)
                raise SystemExit(1)
        elif not self.is_interactive:
            # XXX: this maybe somewhat redundant with validation
            _logger.error(_(
                'Non-interactive session without test plan specified in the '
                'launcher!'))
            raise SystemExit(1)
        else:
            tp_id = self._interactively_pick_test_plan()
            if tp_id is None:
                raise SystemExit(_("No test plan selected."))
        self.ctx.sa.select_test_plan(tp_id)
        description = self.ctx.args.message or self.launcher.session_desc
        self.ctx.sa.update_app_blob(json.dumps(
            {'testplan_id': tp_id,
             'description': description}).encode("UTF-8"))
        bs_jobs = self.ctx.sa.get_bootstrap_todo_list()
        self._run_bootstrap_jobs(bs_jobs)
        self.ctx.sa.finish_bootstrap()

    def _delete_old_sessions(self, ids):
        completed_ids = [s[0] for s in self.ctx.sa.get_old_sessions()]
        self.ctx.sa.delete_sessions(completed_ids + ids)

    def _interactively_pick_test_plan(self):
        test_plan_ids = self.ctx.sa.get_test_plans()
        filtered_tp_ids = set()
        for filter in self.launcher.test_plan_filters:
            filtered_tp_ids.update(fnmatch.filter(test_plan_ids, filter))
        tp_info_list = self._generate_tp_infos(filtered_tp_ids)
        if not tp_info_list:
            print(self.C.RED(_("There were no test plans to select from!")))
            return
        selected_tp = TestPlanBrowser(
            _("Select test plan"), tp_info_list,
            self.launcher.test_plan_default_selection).run()
        return selected_tp

    def _strtobool(self, val):
        return val.lower() in ('y', 'yes', 't', 'true', 'on', '1')

    def _pick_jobs_to_run(self):
        if self.launcher.test_selection_forced:
            if self.launcher.manifest is not Unset:
                self.ctx.sa.save_manifest(
                    {manifest_id:
                     self._strtobool(self.launcher.manifest[manifest_id]) for
                     manifest_id in self.launcher.manifest}
                )
            # by default all tests are selected; so we're done here
            return
        job_list = [self.ctx.sa.get_job(job_id) for job_id in
                    self.ctx.sa.get_static_todo_list()]
        if not job_list:
            print(self.C.RED(_("There were no tests to select from!")))
            return
        test_info_list = self._generate_job_infos(job_list)
        wanted_set = CategoryBrowser(
            _("Choose tests to run on your system:"), test_info_list).run()
        manifest_repr = self.ctx.sa.get_manifest_repr()
        if manifest_repr:
            manifest_answers = ManifestBrowser(
                "System Manifest:", manifest_repr).run()
            self.ctx.sa.save_manifest(manifest_answers)
        # no need to set an alternate selection if the job list not changed
        if len(test_info_list) == len(wanted_set):
            return
        # NOTE: tree.selection is correct but ordered badly. To retain
        # the original ordering we should just treat it as a mask and
        # use it to filter jobs from get_static_todo_list.
        job_id_list = [job_id for job_id in self.ctx.sa.get_static_todo_list()
                       if job_id in wanted_set]
        self.ctx.sa.use_alternate_selection(job_id_list)

    def _handle_last_job_after_resume(self, last_job):
        if last_job is None:
            return
        if self.ctx.args.session_id:
            # session_id is present only if auto-resume is used
            result_dict = {
                'outcome': IJobResult.OUTCOME_PASS,
                'comments': _("Automatically passed after resuming execution"),
            }
            session_share = WellKnownDirsHelper.session_share(
                self.ctx.sa.get_session_id())
            result_path = os.path.join(session_share, '__result')
            if os.path.exists(result_path):
                try:
                    with open(result_path, 'rt') as f:
                        result_dict = json.load(f)
                        # the only really important field in the result is
                        # 'outcome' so let's make sure it doesn't contain
                        # anything stupid
                        if result_dict.get('outcome') not in [
                                'pass', 'fail', 'skip']:
                            result_dict['outcome'] = IJobResult.OUTCOME_PASS
                except json.JSONDecodeError as e:
                    pass
            print(_("Automatically resuming session. "
                    "Outcome of the previous job: {}".format(
                        result_dict['outcome'])))
            result = MemoryJobResult(result_dict)
            self.ctx.sa.use_job_result(last_job, result)
            return

        print(_("Previous session run tried to execute job: {}").format(
            last_job))
        cmd = self._pick_action_cmd([
            Action('s', _("skip that job"), 'skip'),
            Action('p', _("mark it as passed and continue"), 'pass'),
            Action('f', _("mark it as failed and continue"), 'fail'),
            Action('r', _("run it again"), 'run'),
        ], _("What do you want to do with that job?"))
        if cmd == 'skip' or cmd is None:
            result = MemoryJobResult({
                'outcome': IJobResult.OUTCOME_SKIP,
                'comments': _("Skipped after resuming execution")
            })
        elif cmd == 'pass':
            result = MemoryJobResult({
                'outcome': IJobResult.OUTCOME_PASS,
                'comments': _("Passed after resuming execution")
            })
        elif cmd == 'fail':
            result = MemoryJobResult({
                'outcome': IJobResult.OUTCOME_FAIL,
                'comments': _("Failed after resuming execution")
            })
        elif cmd == 'run':
            result = None
        if result:
            self.ctx.sa.use_job_result(last_job, result)

    def _maybe_auto_rerun_jobs(self):
        rerun_candidates = self.ctx.sa.get_rerun_candidates('auto')
        # bail-out early if no job qualifies for rerunning
        if not rerun_candidates:
            return False
        # we wait before retrying
        delay = self.launcher.delay_before_retry
        _logger.info(_("Waiting {} seconds before retrying failed"
                       " jobs...".format(delay)))
        time.sleep(delay)
        candidates = self.ctx.sa.prepare_rerun_candidates(rerun_candidates)
        self._run_jobs(candidates)
        return True

    def _maybe_rerun_jobs(self):
        # create a list of jobs that qualify for rerunning
        rerun_candidates = self.ctx.sa.get_rerun_candidates('manual')
        # bail-out early if no job qualifies for rerunning
        if not rerun_candidates:
            return False
        test_info_list = self._generate_job_infos(rerun_candidates)
        wanted_set = ReRunBrowser(
            _("Select jobs to re-run"), test_info_list, rerun_candidates).run()
        if not wanted_set:
            # nothing selected - nothing to run
            return False
        rerun_candidates = [
            self.ctx.sa.get_job(job_id) for job_id in wanted_set]
        rerun_candidates = self.ctx.sa.prepare_rerun_candidates(
            rerun_candidates)
        # include resource jobs that selected jobs depend on
        self._run_jobs(rerun_candidates)
        return True

    def _get_ui_for_job(self, job):
        class CheckboxUI(NormalUI):
            def considering_job(self, job, job_state):
                pass
        show_out = True
        if self.launcher.output == 'hide-resource-and-attachment':
            if job.plugin in ('local', 'resource', 'attachment'):
                show_out = False
        elif self.launcher.output in ['hide', 'hide-automated']:
            if job.plugin in ('shell', 'local', 'resource', 'attachment'):
                show_out = False
        if 'suppress-output' in job.get_flag_set():
            show_out = False
        if 'use-chunked-io' in job.get_flag_set():
            show_out = True
        if self.ctx.args.dont_suppress_output:
            show_out = True
        return CheckboxUI(self.C.c, show_cmd_output=show_out)

    def register_arguments(self, parser):
        parser.add_argument(
            'launcher', metavar=_('LAUNCHER'), nargs='?',
            help=_('launcher definition file to use'))
        parser.add_argument(
            '--resume', dest='session_id', metavar='SESSION_ID',
            help=SUPPRESS)
        parser.add_argument(
            '--verify', action='store_true',
            help=_('only validate the launcher'))
        parser.add_argument(
            '--title', action='store', metavar='SESSION_NAME',
            help=_('title of the session to use'))
        parser.add_argument(
            "-m", "--message",
            help=_("submission description"))
        parser.add_argument(
            '--dont-suppress-output', action='store_true', default=False,
            help=_('Absolutely always show command output'))
        # the next to options are and should be exact copies of what the
        # top-level command offers - this is here so when someone launches
        # checkbox-cli through launcher, they have those options available
        parser.add_argument('-v', '--verbose', action='store_true', help=_(
            'print more logging from checkbox'))
        parser.add_argument('--debug', action='store_true', help=_(
            'print debug messages from checkbox'))
        parser.add_argument('--clear-cache', action='store_true', help=_(
            'remove cached results from the system'))
        parser.add_argument('--clear-old-sessions', action='store_true', help=_(
            "remove previous sessions' data"))
        parser.add_argument('--version', action='store_true', help=_(
            "show program's version information and exit"))


class CheckboxUI(NormalUI):

    def considering_job(self, job, job_state):
        pass


class Run(MainLoopStage):
    def register_arguments(self, parser):
        parser.add_argument(
            'PATTERN', nargs="*",
            help=_("run jobs matching the given regular expression"))
        parser.add_argument(
            '--non-interactive', action='store_true',
            help=_("skip tests that require interactivity"))
        parser.add_argument(
            '-f', '--output-format',
            default='com.canonical.plainbox::text',
            metavar=_('FORMAT'),
            help=_('save test results in the specified FORMAT'
                   ' (pass ? for a list of choices)'))
        parser.add_argument(
            '-p', '--output-options', default='',
            metavar=_('OPTIONS'),
            help=_('comma-separated list of options for the export mechanism'
                   ' (pass ? for a list of choices)'))
        parser.add_argument(
            '-o', '--output-file', default='-',
            metavar=_('FILE'),  # type=FileType("wb"),
            help=_('save test results to the specified FILE'
                   ' (or to stdout if FILE is -)'))
        parser.add_argument(
            '-t', '--transport',
            metavar=_('TRANSPORT'),
            choices=[_('?')] + list(get_all_transports().keys()),
            help=_('use TRANSPORT to send results somewhere'
                   ' (pass ? for a list of choices)'))
        parser.add_argument(
            '--transport-where',
            metavar=_('WHERE'),
            help=_('where to send data using the selected transport'))
        parser.add_argument(
            '--transport-options',
            metavar=_('OPTIONS'),
            help=_('comma-separated list of key-value options (k=v) to '
                   'be passed to the transport'))
        parser.add_argument(
            '--title', action='store', metavar='SESSION_NAME',
            help=_('title of the session to use'))
        parser.add_argument(
            "-m", "--message",
            help=_("submission description"))

    @property
    def C(self):
        return self._C

    @property
    def sa(self):
        return self.ctx.sa

    @property
    def is_interactive(self):
        """
        Flag indicating that this is an interactive invocation.

        We can then interact with the user when we encounter OUTCOME_UNDECIDED.
        """
        return (sys.stdin.isatty() and sys.stdout.isatty() and not
                self.ctx.args.non_interactive)

    def invoked(self, ctx):
        try:
            self._C = Colorizer()
            self.ctx = ctx

            self._configure_restart()
            config = load_configs()
            self.sa.use_alternate_configuration(config)
            self.sa.start_new_session(
                self.ctx.args.title or 'checkbox-run',
                UnifiedRunner)
            tps = self.sa.get_test_plans()
            self._configure_report()
            selection = ctx.args.PATTERN
            submission_message = self.ctx.args.message
            if len(selection) == 1 and selection[0] in tps:
                self.ctx.sa.update_app_blob(json.dumps(
                    {'testplan_id': selection[0],
                     'description': submission_message}).encode("UTF-8"))
                self.just_run_test_plan(selection[0])
            else:
                self.ctx.sa.update_app_blob(json.dumps(
                    {'description': submission_message}).encode("UTF-8"))
                self.sa.hand_pick_jobs(selection)
                print(self.C.header(_("Running Selected Jobs")))
                self._run_jobs(self.sa.get_dynamic_todo_list())
                # there might have been new jobs instantiated
                while True:
                    self.sa.hand_pick_jobs(ctx.args.PATTERN)
                    todos = self.sa.get_dynamic_todo_list()
                    if not todos:
                        break
                    self._run_jobs(self.sa.get_dynamic_todo_list())
            self.sa.finalize_session()
            self._print_results()
            return 0 if self.sa.get_summary()['fail'] == 0 else 1
        except KeyboardInterrupt:
            return 1

    def just_run_test_plan(self, tp_id):
        self.sa.select_test_plan(tp_id)
        self.sa.bootstrap()
        print(self.C.header(_("Running Selected Test Plan")))
        self._run_jobs(self.sa.get_dynamic_todo_list())

    def _configure_report(self):
        """Configure transport and exporter."""
        if self.ctx.args.output_format == '?':
            print_objs('exporter', self.ctx.sa)
            raise SystemExit(0)
        if self.ctx.args.transport == '?':
            print(', '.join(get_all_transports()))
            raise SystemExit(0)
        if not self.ctx.args.transport:
            if self.ctx.args.transport_where:
                _logger.error(_(
                    "--transport-where is useless without --transport"))
                raise SystemExit(1)
            if self.ctx.args.transport_options:
                _logger.error(_(
                    "--transport-options is useless without --transport"))
                raise SystemExit(1)
            if self.ctx.args.output_file != '-':
                self.transport = 'file'
                self.transport_where = self.ctx.args.output_file
                self.transport_options = ''
            else:
                self.transport = 'stream'
                self.transport_where = 'stdout'
                self.transport_options = ''
        else:
            if self.ctx.args.transport not in get_all_transports():
                _logger.error("The selected transport %r is not available",
                              self.ctx.args.transport)
                raise SystemExit(1)
            self.transport = self.ctx.args.transport
            self.transport_where = self.ctx.args.transport_where
            self.transport_options = self.ctx.args.transport_options
        self.exporter = self.ctx.args.output_format
        self.exporter_opts = self.ctx.args.output_options

    def _print_results(self):
        all_transports = get_all_transports()
        transport = get_all_transports()[self.transport](
            self.transport_where, self.transport_options)
        print(self.C.header(_("Results")))
        if self.transport == 'file':
            print(_("Saving results to {}").format(self.transport_where))
        elif self.transport == 'certification':
            print(_("Sending results to {}").format(self.transport_where))
        self.sa.export_to_transport(
            self.exporter, transport, self.exporter_opts)

    def _configure_restart(self):
        strategy = detect_restart_strategy(session_type='local')
        snap_name = os.getenv('SNAP_NAME')
        if snap_name:
            # NOTE: This implies that any snap wishing to include a
            # Checkbox snap to be autostarted creates a snapcraft
            # app called "checkbox-cli"
            respawn_cmd = '/snap/bin/{}.checkbox-cli'.format(snap_name)
        else:
            respawn_cmd = sys.argv[0]  # entry-point to checkbox
        respawn_cmd += ' --resume {}'  # interpolate with session_id
        self.sa.configure_application_restart(
            lambda session_id: [respawn_cmd.format(session_id)])


class List():
    def register_arguments(self, parser):
        parser.add_argument(
            'GROUP', nargs='?',
            help=_("list objects from the specified group"))
        parser.add_argument(
            '-a', '--attrs', default=False, action="store_true",
            help=_("show object attributes"))
        parser.add_argument(
            '-f', '--format', type=str,
            help=_(("output format, as passed to print function. "
                    "Use '?' to list possible values")))

    def invoked(self, ctx):
        if ctx.args.GROUP == 'all-jobs':
            if ctx.args.attrs:
                print_objs('job', ctx.sa, True)

                def filter_fun(u): return u.attrs['template_unit'] == 'job'
                print_objs('template', ctx.sa, True, filter_fun)
            jobs = get_all_jobs(ctx.sa)
            if ctx.args.format == '?':
                all_keys = set()
                for job in jobs:
                    all_keys.update(job.keys())
                print(_('Available fields are:'))
                print(', '.join(sorted(list(all_keys))))
                return
            if not ctx.args.format:
                # setting default in parser.add_argument would apply to all
                # the list invocations. We want default to be present only for
                # the 'all-jobs' group.
                ctx.args.format = 'id: {full_id}\n{_summary}\n'
            for job in jobs:
                unescaped = ctx.args.format.replace(
                    '\\n', '\n').replace('\\t', '\t')

                class DefaultKeyedDict(defaultdict):
                    def __missing__(self, key):
                        return _('<missing {}>').format(key)
                # formatters are allowed to use special field 'unit_type' so
                # let's add it to the job representation
                assert 'unit_type' not in job.keys()
                if job.get('template_unit') == 'job':
                    job['unit_type'] = 'template_job'
                else:
                    job['unit_type'] = 'job'
                print(Formatter().vformat(
                    unescaped, (), DefaultKeyedDict(None, job)), end='')
            return
        elif ctx.args.format:
            print(_("--format applies only to 'all-jobs' group.  Ignoring..."))
        print_objs(ctx.args.GROUP, ctx.sa, ctx.args.attrs)


class ListBootstrapped():
    @property
    def sa(self):
        return self.ctx.sa

    def register_arguments(self, parser):
        parser.add_argument(
            'TEST_PLAN',
            help=_("test-plan id to bootstrap"))
        parser.add_argument(
            '-f', '--format', type=str, default="{full_id}\n",
            help=_(("output format, as passed to print function. "
                    "Use '?' to list possible values")))

    def invoked(self, ctx):
        self.ctx = ctx
        self.sa.start_new_session('checkbox-listing-ephemeral')
        tps = self.sa.get_test_plans()
        if ctx.args.TEST_PLAN not in tps:
            raise SystemExit('Test plan not found')
        self.sa.select_test_plan(ctx.args.TEST_PLAN)
        self.sa.bootstrap()
        jobs = []
        for job in self.sa.get_static_todo_list():
            job_unit = self.sa.get_job(job)
            attrs = job_unit._raw_data.copy()
            attrs['full_id'] = job_unit.id
            attrs['id'] = job_unit.partial_id
            jobs.append(attrs)
        if ctx.args.format == '?':
            all_keys = set()
            for job in jobs:
                all_keys.update(job.keys())
            print(_('Available fields are:'))
            print(', '.join(sorted(list(all_keys))))
            return
        if ctx.args.format:
            for job in jobs:
                unescaped = ctx.args.format.replace(
                    '\\n', '\n').replace('\\t', '\t')

                class DefaultKeyedDict(defaultdict):
                    def __missing__(self, key):
                        return _('<missing {}>').format(key)
                print(Formatter().vformat(
                    unescaped, (), DefaultKeyedDict(None, job)), end='')
        else:
            for job_id in jobs:
                print(job_id)


class TestPlanExport():

    @property
    def sa(self):
        return self.ctx.sa

    def register_arguments(self, parser):
        parser.add_argument(
            'TEST_PLAN',
            help=_("test-plan id to bootstrap"))
        parser.add_argument(
            '-n', '--nofake', action='store_true')

    def invoked(self, ctx):
        self.ctx = ctx
        if ctx.args.nofake:
            self.sa.start_new_session('tp-export-ephemeral')
        else:
            from plainbox.impl.execution import FakeJobRunner
            self.sa.start_new_session('tp-export-ephemeral', FakeJobRunner)
            self.sa._context.state._fake_resources = True
        tps = self.sa.get_test_plans()
        if ctx.args.TEST_PLAN not in tps:
            raise SystemExit('Test plan not found')
        self.sa.select_test_plan(ctx.args.TEST_PLAN)
        self.sa.bootstrap()
        path = self.sa.export_to_file(
            'com.canonical.plainbox::tp-export', [],
            self.sa._manager.storage.location,
            slugify(self.sa._manager.test_plans[0].name))
        print(path)


def get_all_jobs(sa):
    providers = sa.get_selected_providers()
    root = Explorer(providers).get_object_tree()

    def get_jobs(obj):
        jobs = []
        if obj.group == 'job' or (
                obj.group == 'template' and obj.attrs['template_unit'] == 'job'):
            attrs = dict(obj._impl._raw_data.copy())
            attrs['full_id'] = obj.name
            jobs.append(attrs)
        for child in obj.children:
            jobs += get_jobs(child)
        return jobs
    return sorted(get_jobs(root), key=operator.itemgetter('full_id'))


def print_objs(group, sa, show_attrs=False, filter_fun=None):
    providers = sa.get_selected_providers()
    obj = Explorer(providers).get_object_tree()

    def _show(obj, indent):
        if group is None or obj.group == group:
            # object must satisfy filter_fun (if supplied) to be printed
            if filter_fun and not filter_fun(obj):
                return
            # Display the object name and group
            print("{}{} {!r}".format(indent, obj.group, obj.name))
            indent += "  "
            if show_attrs:
                for key, value in obj.attrs.items():
                    print("{}{:15}: {!r}".format(indent, key, value))
        if obj.children:
            if group is None:
                print("{}{}".format(indent, _("children")))
                indent += "  "
            for child in obj.children:
                _show(child, indent)

    _show(obj, "")



# This file is part of Checkbox.
#
# Copyright 2022 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
Hacks to make testing of argparse prints work on python <= 3.9 and >= 3.10.
"""

import argparse


def optionals_section():
    """
    Return a string that's the default title for the group of optional
    arguments, as returned by argparse.
    """
    try:
        # Let's try pulling the info directly from the ArgumentParser object.
        # This means peekign at private parts of it.
        return argparse.ArgumentParser()._optionals.title
    except Exception:
        # If anything goes south we need to guess the string basing on the
        # pythoon version
        import sys
        vi = sys.version_info
        if vi.major >= 3 and vi.minor >= 10:
            return 'options'
        else:
            return 'optional arguments'


optionals_section = optionals_section()



# This file is part of Checkbox.
#
# Copyright 2013 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`plainbox.impl.ctrl` -- Controller Classes
===============================================

Session controller classes implement the glue between models (jobs, test plans,
session state) and the rest of the application. They encapsulate knowledge that
used to be special-cased and sprinkled around various parts of both plainbox
and particular plainbox-using applications.

Execution controllers are used by the :class:`~plainbox.impl.runner.JobRunner`
class to select the best method to execute a command of a particular job.  This
is mostly applicable to jobs that need to run as another user, typically as
root, as the method that is used to effectively gain root differs depending on
circumstances.
"""

import abc
import contextlib
import errno
try:
    import grp
except ImportError:
    grp = None
import itertools
import json
import logging
import os
import tempfile
import subprocess
import sys
import threading
from subprocess import check_output, CalledProcessError, STDOUT

from plainbox.abc import IJobResult
from plainbox.abc import ISessionStateController
from plainbox.i18n import gettext as _
from plainbox.impl import get_plainbox_dir
from plainbox.impl.depmgr import DependencyDuplicateError
from plainbox.impl.depmgr import DependencyMissingError
from plainbox.impl.resource import ExpressionCannotEvaluateError
from plainbox.impl.resource import ExpressionFailedError
from plainbox.impl.resource import ResourceProgramError
from plainbox.impl.resource import Resource
from plainbox.impl.secure.config import Unset
from plainbox.impl.secure.origin import JobOutputTextSource
from plainbox.impl.secure.providers.v1 import Provider1
from plainbox.impl.secure.rfc822 import RFC822SyntaxError
from plainbox.impl.secure.rfc822 import gen_rfc822_records
from plainbox.impl.session.jobs import InhibitionCause
from plainbox.impl.session.jobs import JobReadinessInhibitor
from plainbox.impl.unit.job import JobDefinition
from plainbox.impl.unit.template import TemplateUnit
from plainbox.impl.unit.unit import MissingParam
from plainbox.impl.validation import Severity
from plainbox.vendor import morris
from plainbox.vendor import extcmd

__all__ = [
    'CheckBoxSessionStateController',
    'checkbox_session_state_ctrl',
]


logger = logging.getLogger("plainbox.ctrl")


class CheckBoxSessionStateController(ISessionStateController):
    """
    A combo controller for CheckBox-like jobs.

    This controller implements the following features:

        * A job may depend on another job, this is expressed via the 'depends'
          attribute. Cyclic dependencies are not allowed. A job will become
          inhibited if any of its dependencies have outcome other than
          OUTCOME_PASS
        * A job may require that a particular resource expression evaluates to
          true. This is expressed via the 'requires' attribute. A job will
          become inhibited if any of the requirement programs evaluates to
          value other than True.
        * A job may have the attribute 'plugin' equal to "resource" which will
          cause the controller to interpret the stdout of the command as a set
          of resource definitions.
    """

    def get_dependency_set(self, job):
        """
        Get the set of direct dependencies of a particular job.

        :param job:
            A IJobDefinition instance that is to be visited
        :returns:
            set of pairs (dep_type, job_id)

        Returns a set of pairs (dep_type, job_id) that describe all
        dependencies of the specified job. The first element in the pair,
        dep_type, is either DEP_TYPE_DIRECT, DEP_TYPE_ORDERING or
        DEP_TYPE_RESOURCE. The second element is the id of the job.
        """
        direct = DependencyMissingError.DEP_TYPE_DIRECT
        ordering = DependencyMissingError.DEP_TYPE_ORDERING
        resource = DependencyMissingError.DEP_TYPE_RESOURCE
        direct_deps = job.get_direct_dependencies()
        after_deps = job.get_after_dependencies()
        try:
            resource_deps = job.get_resource_dependencies()
        except ResourceProgramError:
            resource_deps = ()
        result = set(itertools.chain(
            zip(itertools.repeat(direct), direct_deps),
            zip(itertools.repeat(resource), resource_deps),
            zip(itertools.repeat(ordering), after_deps)))
        return result

    def get_inhibitor_list(self, session_state, job):
        """
        Get a list of readiness inhibitors that inhibit a particular job.

        :param session_state:
            A SessionState instance that is used to interrogate the
            state of the session where it matters for a particular
            job. Currently this is used to access resources and job
            results.
        :param job:
            A JobDefinition instance
        :returns:
            List of JobReadinessInhibitor
        """
        inhibitors = []
        # Check if all job resource requirements are met
        prog = job.get_resource_program()
        if prog is not None:
            try:
                prog.evaluate_or_raise(session_state.resource_map)
            except ExpressionCannotEvaluateError as exc:
                for resource_id in exc.expression.resource_id_list:
                    if session_state.job_state_map[resource_id].result.outcome == 'pass':
                        continue
                    # Lookup the related job (the job that provides the
                    # resources needed by the expression that cannot be
                    # evaluated)
                    related_job = session_state.job_state_map[resource_id].job
                    # Add A PENDING_RESOURCE inhibitor as we are unable to
                    # determine if the resource requirement is met or not. This
                    # can happen if the resource job did not ran for any reason
                    # (it can either be prevented from running by normal means
                    # or simply be on the run_list but just was not executed
                    # yet).
                    inhibitor = JobReadinessInhibitor(
                        cause=InhibitionCause.PENDING_RESOURCE,
                        related_job=related_job,
                        related_expression=exc.expression)
                    inhibitors.append(inhibitor)
            except ExpressionFailedError as exc:
                # When expressions fail then all the associated resources are
                # marked as failed since we don't want to get into the analysis
                # of logic expressions to know any "better".
                for resource_id in exc.expression.resource_id_list:
                    # Lookup the related job (the job that provides the
                    # resources needed by the expression that failed)
                    related_job = session_state.job_state_map[resource_id].job
                    # Add a FAILED_RESOURCE inhibitor as we have all the data
                    # to run the requirement program but it simply returns a
                    # non-True value. This typically indicates a missing
                    # software package or necessary hardware.
                    inhibitor = JobReadinessInhibitor(
                        cause=InhibitionCause.FAILED_RESOURCE,
                        related_job=related_job,
                        related_expression=exc.expression)
                    inhibitors.append(inhibitor)
        # Check if all job dependencies ran successfully
        for dep_id in sorted(job.get_direct_dependencies()):
            dep_job_state = session_state.job_state_map[dep_id]
            # If the dependency did not have a chance to run yet add the
            # PENDING_DEP inhibitor.
            if dep_job_state.result.outcome == IJobResult.OUTCOME_NONE:
                inhibitor = JobReadinessInhibitor(
                    cause=InhibitionCause.PENDING_DEP,
                    related_job=dep_job_state.job)
                inhibitors.append(inhibitor)
            # If the dependency is anything but successful add the
            # FAILED_DEP inhibitor. In theory the PENDING_DEP code above
            # could be discarded but this would loose context and would
            # prevent the operator from actually understanding why a job
            # cannot run.
            elif dep_job_state.result.outcome != IJobResult.OUTCOME_PASS:
                inhibitor = JobReadinessInhibitor(
                    cause=InhibitionCause.FAILED_DEP,
                    related_job=dep_job_state.job)
                inhibitors.append(inhibitor)
        # Check if all "after" dependencies ran yet
        for dep_id in sorted(job.get_after_dependencies()):
            dep_job_state = session_state.job_state_map[dep_id]
            # If the dependency did not have a chance to run yet add the
            # PENDING_DEP inhibitor.
            if dep_job_state.result.outcome == IJobResult.OUTCOME_NONE:
                inhibitor = JobReadinessInhibitor(
                    cause=InhibitionCause.PENDING_DEP,
                    related_job=dep_job_state.job)
                inhibitors.append(inhibitor)
        for dep_id in sorted(job.get_salvage_dependencies()):
            dep_job_state = session_state.job_state_map[dep_id]
            if dep_job_state.result.outcome != IJobResult.OUTCOME_FAIL:
                inhibitor = JobReadinessInhibitor(
                    cause=InhibitionCause.NOT_FAILED_DEP,
                    related_job=dep_job_state.job)
                inhibitors.append(inhibitor)
        return inhibitors

    def observe_result(self, session_state, job, result,
                       fake_resources=False):
        """
        Notice the specified test result and update readiness state.

        :param session_state:
            A SessionState object
        :param job:
            A JobDefinition object
        :param result:
            A IJobResult object
        :param fake_resources:
            An optional parameter to trigger test plan export execution mode
            using fake resourceobjects

        This function updates the internal result collection with the data from
        the specified test result. Results can safely override older results.
        Results also change the ready map (jobs that can run) because of
        dependency relations.

        Some results have deeper meaning, those are results for resource jobs.
        They are discussed in detail below:

        Resource jobs produce resource records which are used as data to run
        requirement expressions against. Each time a result for a resource job
        is presented to the session it will be parsed as a collection of RFC822
        records. A new entry is created in the resource map (entirely replacing
        any old entries), with a list of the resources that were parsed from
        the IO log.
        """
        # Store the result in job_state_map
        session_state.job_state_map[job.id].result = result
        session_state.on_job_state_map_changed()
        session_state.on_job_result_changed(job, result)
        # Treat some jobs specially and interpret their output
        if job.plugin == "resource":
            self._process_resource_result(
                session_state, job, result, fake_resources)

    def _process_resource_result(self, session_state, job, result,
                                 fake_resources=False):
        """
        Analyze a result of a CheckBox "resource" job and generate
        or replace resource records.
        """
        self._parse_and_store_resource(session_state, job, result)
        if session_state.resource_map[job.id] != [Resource({})]:
            self._instantiate_templates(
                session_state, job, result, fake_resources)

    def _parse_and_store_resource(self, session_state, job, result):
        # NOTE: https://bugs.launchpad.net/checkbox/+bug/1297928
        # If we are resuming from a session that had a resource job that
        # never ran, we will see an empty MemoryJobResult object.
        # Processing empty I/O log would create an empty resource list
        # and that state is different from the state the session started
        # before it was suspended, so don't
        if result.outcome is IJobResult.OUTCOME_NONE:
            return
        new_resource_list = []
        for record in gen_rfc822_records_from_io_log(job, result):
            # XXX: Consider forwarding the origin object here.  I guess we
            # should have from_frc822_record as with JobDefinition
            resource = Resource(record.data)
            logger.debug(
                _("Storing resource record %r: %s"), job.id, resource)
            new_resource_list.append(resource)
        # Create an empty resource object to properly fail __getattr__ calls
        if not new_resource_list:
            new_resource_list = [Resource({})]
        # Replace any old resources with the new resource list
        session_state.set_resource_list(job.id, new_resource_list)

    def _instantiate_templates(self, session_state, job, result,
                               fake_resources=False):
        # NOTE: https://bugs.launchpad.net/checkbox/+bug/1297928
        # If we are resuming from a session that had a resource job that
        # never ran, we will see an empty MemoryJobResult object.
        # Processing empty I/O log would create an empty resource list
        # and that state is different from the state the session started
        # before it was suspended, so don't
        if result.outcome is IJobResult.OUTCOME_NONE:
            return
        for unit in session_state.unit_list:
            if isinstance(unit, TemplateUnit) and unit.resource_id == job.id:
                logger.info(_("Instantiating unit: %s"), unit)
                for new_unit in unit.instantiate_all(
                        session_state.resource_map[job.id], fake_resources):
                    try:
                        check_result = new_unit.check()
                    except MissingParam as m:
                        logger.debug(_("Ignoring %s with missing "
                                         "template parameter %s"),
                                         new_unit._raw_data.get('id'),
                                         m.parameter)
                        continue
                    # Only ignore jobs for which check() returns an error
                    if [c for c in check_result
                            if c.severity == Severity.error]:
                        logger.error(_("Ignoring invalid generated job %s"),
                                     new_unit.id)
                    else:
                        session_state.add_unit(new_unit, via=job, recompute=False)
        session_state._recompute_job_readiness()


def gen_rfc822_records_from_io_log(job, result):
    """
    Convert io_log from a job result to a sequence of rfc822 records
    """
    logger.debug(_("processing output from a job: %r"), job)
    # Select all stdout lines from the io log
    line_gen = (record[2].decode('UTF-8', errors='replace')
                for record in result.get_io_log()
                if record[1] == 'stdout')
    # Allow the generated records to be traced back to the job that defined
    # the command which produced (printed) them.
    source = JobOutputTextSource(job)
    try:
        # Parse rfc822 records from the subsequent lines
        for record in gen_rfc822_records(line_gen, source=source):
            yield record
    except RFC822SyntaxError as exc:
        # When this exception happens we will _still_ store all the
        # preceding records. This is worth testing

        logger.warning(
            # TRANSLATORS: keep the word "local" untranslated. It is a
            # special type of job that needs to be distinguished.
            _("local script %s returned invalid RFC822 data: %s"),
            job.id, exc)


checkbox_session_state_ctrl = CheckBoxSessionStateController()


class SymLinkNest:
    """
    A class for setting up a control directory with symlinked executables
    """

    def __init__(self, dirname):
        self._dirname = dirname

    def add_provider(self, provider):
        """
        Add all of the executables associated a particular provider

        :param provider:
            A Provider1 instance
        """
        for filename in provider.executable_list:
            self.add_executable(filename)

    def add_executable(self, filename):
        """
        Add a executable to the control directory
        """
        logger.debug(
            _("Adding executable %s to nest %s"),
            filename, self._dirname)
        dest = os.path.join(self._dirname, os.path.basename(filename))
        try:
            os.symlink(filename, dest)
        except OSError as exc:
            # Allow symlinks to fail on Windows where it requires some
            # untold voodoo magic to work (aka running as root)
            logger.error(
                _("Unable to create symlink s%s -> %s: %r"),
                filename, dest, exc)



# This file is part of Checkbox.
#
# Copyright 2017-2019 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
This module contains implementation of the service end of the remote execution
functionality.
"""
import gettext
import logging
import os
import socket
import sys
from plainbox.impl.secure.sudo_broker import is_passwordless_sudo
from plainbox.impl.session.remote_assistant import RemoteSessionAssistant
from plainbox.impl.session.restart import RemoteDebRestartStrategy
from plainbox.impl.session.restart import RemoteSnappyRestartStrategy
from plainbox.vendor import rpyc
from plainbox.vendor.rpyc.utils.server import ThreadedServer

_ = gettext.gettext
_logger = logging.getLogger("service")


class SessionAssistantSlave(rpyc.Service):

    session_assistant = None
    controlling_master_conn = None
    master_blaster = None

    def exposed_get_sa(*args):
        return SessionAssistantSlave.session_assistant

    def exposed_register_master_blaster(self, callable):
        """
        Register a callable that will be called when the slave decides to
        disconnect the master. This should be used to prepare the master for
        the disconnection, so it can differentiate between network failures
        and a planned disconnect.
        The callable will be called with one param - a string with a reason
        for the disconnect.
        """
        SessionAssistantSlave.master_blaster = callable

    def on_connect(self, conn):
        try:
            if SessionAssistantSlave.master_blaster:
                msg = 'Forcefully disconnected by new master from {}:{}'.format(
                    conn._config['endpoints'][1][0], conn._config['endpoints'][1][1])
                SessionAssistantSlave.master_blaster(msg)
                old_master = SessionAssistantSlave.controlling_master_conn
                if old_master is not None:
                    old_master.close()
                SessionAssistantSlave.master_blaster = None

            SessionAssistantSlave.controlling_master_conn = conn
        except TimeoutError as exc:
            # this happens when the reference to .master_blaster times out,
            # meaning the master is blocked on an urwid screen or some other
            # thread blocking operation. In any case it means there was a
            # previous master, so we need to kill it
            old_master = SessionAssistantSlave.controlling_master_conn
            SessionAssistantSlave.master_blaster = None
            old_master.close()
            SessionAssistantSlave.controlling_master_conn = conn

    def on_disconnect(self, conn):
        SessionAssistantSlave.master_blaster = None
        self.controlling_master_conn = None


class RemoteSlave():
    """
    Run checkbox instance as a service

    RemoteSlave implements functionality for the half that's actually running
    the tests - the one that was summoned using `checkbox-cli service`. This
    part should be run on system-under-test.
    """

    name = 'service'

    def invoked(self, ctx):
        if os.geteuid():
            raise SystemExit(_("Checkbox service must be run by root!"))
        if not is_passwordless_sudo():
            raise SystemExit(
                _("System is not configured to run sudo without a password!"))
        slave_port = ctx.args.port

        # Check if able to connect to the slave port as indicator of there
        # already being a slave running
        def slave_port_open():
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(0.5)
            result = sock.connect_ex(('127.0.0.1', slave_port))
            sock.close()
            return result
        if slave_port_open() == 0:
            raise SystemExit(_("Found port {} is open. Is Checkbox service"
                               " already running?").format(slave_port))

        SessionAssistantSlave.session_assistant = RemoteSessionAssistant(
            lambda s: [sys.argv[0] + 'service'])
        snap_data = os.getenv('SNAP_DATA')
        snap_rev = os.getenv('SNAP_REVISION')
        remote_restart_strategy_debug = os.getenv('REMOTE_RESTART_DEBUG')
        if (snap_data and snap_rev) or ctx.args.resume:
            if remote_restart_strategy_debug:
                strategy = RemoteSnappyRestartStrategy(debug=True)
            else:
                strategy = RemoteSnappyRestartStrategy()
            if os.path.exists(strategy.session_resume_filename):
                with open(strategy.session_resume_filename, 'rt') as f:
                    session_id = f.readline()
                SessionAssistantSlave.session_assistant.resume_by_id(
                    session_id)
            elif ctx.args.resume:
                # XXX: explicitly passing None to not have to bump Remote API
                # TODO: remove on the next Remote API bump
                SessionAssistantSlave.session_assistant.resume_by_id(None)
        else:
            _logger.info("RemoteDebRestartStrategy")
            if remote_restart_strategy_debug:
                strategy = RemoteDebRestartStrategy(debug=True)
            else:
                strategy = RemoteDebRestartStrategy()
            if os.path.exists(strategy.session_resume_filename):
                with open(strategy.session_resume_filename, 'rt') as f:
                    session_id = f.readline()
                _logger.info(
                    "RemoteDebRestartStrategy resume_by_id %r", session_id)
                SessionAssistantSlave.session_assistant.resume_by_id(
                    session_id)
        self._server = ThreadedServer(
            SessionAssistantSlave,
            port=slave_port,
            protocol_config={
                "allow_all_attrs": True,
                "allow_setattr": True,
                "sync_request_timeout": 1,
                "propagate_SystemExit_locally": True
            },
        )
        SessionAssistantSlave.session_assistant.terminate_cb = (
            self._server.close)
        self._server.start()

    def register_arguments(self, parser):
        parser.add_argument('--resume', action='store_true', help=_(
            "resume last session"))
        parser.add_argument('--port', type=int, default=18871, help=_(
            "port to listen on"))



# This file is part of Checkbox.
#
# Copyright 2012, 2013 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`plainbox.impl.session.storage` -- storage for sessions
============================================================

This module contains storage support code for handling sessions. Each location
is wrapped by a :class:`SessionStorage` instance. That latter class be used to
create (allocate) and remove all of the files associated with a particular
session.
"""

import datetime
import errno
import logging
import os
import shutil
import stat

from plainbox.i18n import gettext as _, ngettext
from plainbox.impl.runner import slugify

logger = logging.getLogger("plainbox.session.storage")


class WellKnownDirsHelper():
    """
    Helper class that knows about well known directories and paths.

    This class simply gets rid of various magic path names that are used during
    an checkbox invocation. The methods :meth:`populate_base()` and
    :meth:`populate_session()` are convenience utilities to create the
    directory structure and ensure permissions are correct.
    """

    base_of_everything = '/var/tmp/checkbox-ng'

    @classmethod
    def populate_base(cls):
        """
        Create all of the well known static directories that are not session
        specific
        """
        oldmask = os.umask(000)
        for dirname in cls._base_directories():
            if not os.path.exists(dirname):
                os.makedirs(dirname)
                # TODO: test that is 0o777 and if not attempt to fix?
        os.umask(oldmask)

    @classmethod
    def populate_session(cls, session_id):
        """
        Create the session specific directories
        """
        oldmask = os.umask(000)
        for dirname in cls._session_directories(session_id):
            if not os.path.exists(dirname):
                os.makedirs(dirname)
        os.umask(oldmask)
        return cls.session_dir(session_id)

    @classmethod
    def _base_directories(cls):
        return [cls.base_of_everything, cls.session_repository()]

    @classmethod
    def _session_directories(cls, session_id):
        return [cls.session_dir(session_id),
                cls.session_share(session_id),
                cls.io_logs(session_id)]

    @classmethod
    def session_repository(cls):
        return os.path.join(cls.base_of_everything, "sessions")

    @classmethod
    def session_dir(cls, session_id):
        return os.path.join(cls.session_repository(),
                            "{}.{}".format(session_id, "session"))

    @classmethod
    def io_logs(cls, session_id):
        return os.path.join(cls.session_dir(session_id), "io-logs")

    @classmethod
    def session_share(cls, session_id):
        return os.path.join(cls.session_dir(session_id), "session-share")

    @classmethod
    def manifest_file(cls):
        return os.path.join(cls.base_of_everything, "machine-manifest.json")

    @classmethod
    def get_storage_list(self):
        """
        Enumerate stored sessions in the repository.

        If the repository directory is not present then an empty list is
        returned.

        :returns:
            list of :class:`SessionStorage` representing discovered sessions
            sorted by their age (youngest first)
        """
        repo = WellKnownDirsHelper().session_repository()
        logger.debug(_("Enumerating sessions in %s"), repo)
        try:
            # Try to enumerate the directory
            item_list = sorted(os.listdir(repo),
                               key=lambda x: os.stat(os.path.join(
                                   repo, x)).st_mtime, reverse=True)
        except OSError as exc:
            # If the directory does not exist,
            # silently return empty collection
            if exc.errno == errno.ENOENT:
                return []
            # Don't silence any other errors
            raise
        session_list = []
        # Check each item by looking for directories
        for item in item_list:
            pathname = os.path.join(repo, item)
            # Make sure not to follow any symlinks here
            stat_result = os.lstat(pathname)
            # Consider non-hidden directories that end with the word .session
            if (not item.startswith(".") and item.endswith(".session")
                    and stat.S_ISDIR(stat_result.st_mode)):
                logger.debug(_("Found possible session in %r"), pathname)
                session = SessionStorage(os.path.splitext(item)[0])
                session_list.append(session)
        # Return the full list
        return session_list


class LockedStorageError(IOError):
    """
    Exception raised when SessionStorage.save_checkpoint() finds an existing
    'next' file from a (presumably) previous call to save_checkpoint() that
    got interrupted
    """


class SessionStorage:
    """
    Abstraction for storage area that is used by :class:`SessionState` to
    keep some persistent and volatile data.

    This class implements functions performing input/output operations
    on session checkpoint data. The location property can be used for keeping
    any additional files or directories but keep in mind that they will
    be removed by :meth:`SessionStorage.remove()`

    This class indirectly collaborates with :class:`SessionSuspendHelper` and
    :class:`SessionResumeHelper`.
    """

    _SESSION_FILE = 'session'

    _SESSION_FILE_NEXT = 'session.next'

    def __init__(self, id):
        """
        Initialize a :class:`SessionStorage` with the given location.

        The location is not created. If you want to ensure that it exists
        call :meth:`create()` instead.
        """
        self._id = id

    def __repr__(self):
        return "<{} location:{!r}>".format(
            self.__class__.__name__, self.location)

    @property
    def location(self):
        """
        location of the session storage
        """
        return WellKnownDirsHelper.session_dir(self.id)

    @property
    def id(self):
        """
        identifier of the session storage (name of the random directory)
        """
        return self._id

    @property
    def session_file(self):
        """
        pathname of the session state file
        """
        return os.path.join(self.location, self._SESSION_FILE)

    @classmethod
    def create(cls, prefix='pbox-'):
        """
        Create a new :class:`SessionStorage` in a subdirectory of the base
        directory. The directory structure will be created if it does not exist
        and will writable by any user.

        :param prefix:
            String which should prefix all session filenames. The prefix is
            sluggified before use.
        """
        WellKnownDirsHelper.populate_base()

        isoformat = "%Y-%m-%dT%H.%M.%S"
        timestamp = datetime.datetime.utcnow().strftime(isoformat)
        session_id = "{prefix}{timestamp}".format(prefix=slugify(prefix),
                                                  timestamp=timestamp)
        uniq = 1
        while os.path.exists(WellKnownDirsHelper.session_dir(session_id)):
            session_id = "{prefix}{timestamp}_({uniq})".format(
                prefix=slugify(prefix), timestamp=timestamp, uniq=uniq)
            uniq += 1
        session_dir = WellKnownDirsHelper.populate_session(session_id)

        logger.debug(_("Created new storage in %r"), session_dir)
        self = cls(session_id)
        return self

    def remove(self):
        """
        Remove all filesystem entries associated with this instance.
        """
        logger.debug(_("Removing session storage from %r"), self.location)

        def error_handler(function, path, excinfo):
            logger.warning(_("Cannot remove %s"), path)
        shutil.rmtree(self.location, onerror=error_handler)

    def load_checkpoint(self):
        """
        Load checkpoint data from the filesystem

        :returns: data from the most recent checkpoint
        :rtype: bytes

        :raises IOError, OSError:
            on various problems related to accessing the filesystem
        """
        # Open the location directory
        location_fd = os.open(self.location, os.O_DIRECTORY)
        try:
            # Open the current session file in the location directory
            session_fd = os.open(
                self._SESSION_FILE, os.O_RDONLY, dir_fd=location_fd)
            # Stat the file to know how much to read
            session_stat = os.fstat(session_fd)
            try:
                # Read session data
                data = os.read(session_fd, session_stat.st_size)
                if len(data) != session_stat.st_size:
                    raise IOError(_("partial read?"))
            finally:
                # Close the session file
                os.close(session_fd)
        except IOError as exc:
            if exc.errno == errno.ENOENT:
                # Treat lack of 'session' file as an empty file
                return b''
            raise
        else:
            return data
        finally:
            # Close the location directory
            os.close(location_fd)

    def save_checkpoint(self, data):
        """
        Save checkpoint data to the filesystem.

        The directory associated with this :class:`SessionStorage` must already
        exist. Typically the instance should be obtained by calling
        :meth:`SessionStorage.create()` which will ensure that this is already
        the case.

        :raises TypeError:
            if data is not a bytes object.

        :raises LockedStorageError:
            if leftovers from previous save_checkpoint() have been detected.
            Normally those should never be here but in certain cases that is
            possible. Callers might want to call :meth:`break_lock()`
            to resolve the problem and try again.

        :raises IOError, OSError:
            on various problems related to accessing the filesystem.
            Typically permission errors may be reported here.
        """
        if not isinstance(data, bytes):
            raise TypeError("data must be bytes")
        logger.debug(ngettext(
            "Saving %d byte of data (%s)",
            "Saving %d bytes of data (%s)",
            len(data)), len(data), "UNIX, python 3.3 or newer")
        # Open the location directory, we need to fsync that later
        # XXX: this may fail, maybe we should keep the fd open all the time?
        location_fd = os.open(self.location, os.O_DIRECTORY)
        logger.debug(
            _("Opened %r as descriptor %d"), self.location, location_fd)
        try:
            # Open the "next" file in the location_directory
            #
            # Use openat(2) to ensure we always open a file relative to the
            # directory we already opened above. This is essential for fsync(2)
            # calls made below.
            #
            # Use "write" + "create" + "exclusive" flags so that no race
            # condition is possible.
            #
            # This will never return -1, it throws IOError when anything is
            # wrong. The caller has to catch this.
            #
            # As a special exception, this code handles EEXISTS
            # (FIleExistsError) and converts that to LockedStorageError
            # that can be especially handled by some layer above.
            try:
                next_session_fd = os.open(
                    self._SESSION_FILE_NEXT,
                    os.O_WRONLY | os.O_CREAT | os.O_EXCL, 0o644,
                    dir_fd=location_fd)
            except FileExistsError:
                raise LockedStorageError()
            logger.debug(
                _("Opened next session file %s as descriptor %d"),
                self._SESSION_FILE_NEXT, next_session_fd)
            try:
                # Write session data to disk
                #
                # I cannot find conclusive evidence but it seems that
                # os.write() handles partial writes internally. In case we do
                # get a partial write _or_ we run out of disk space, raise an
                # explicit IOError.
                num_written = os.write(next_session_fd, data)
                logger.debug(ngettext(
                    "Wrote %d byte of data to descriptor %d",
                    "Wrote %d bytes of data to descriptor %d", num_written),
                    num_written, next_session_fd)
                if num_written != len(data):
                    raise IOError(_("partial write?"))
            except Exception as exc:
                logger.warning(_("Unable to complete write: %r"), exc)
                # If anything goes wrong we should unlink the next file. As
                # with the open() call above we use unlinkat to prevent race
                # conditions.
                # TRANSLATORS: unlinking as in deleting a file
                logger.warning(_("Unlinking %r"), self._SESSION_FILE_NEXT)
                os.unlink(self._SESSION_FILE_NEXT, dir_fd=location_fd)
            else:
                # If the write was successful we must flush kernel buffers.
                #
                # We want to be sure this data is really on disk by now as we
                # may crash the machine soon after this method exits.
                logger.debug(
                    # TRANSLATORS: please don't translate fsync()
                    _("Calling fsync() on descriptor %d"), next_session_fd)
                try:
                    os.fsync(next_session_fd)
                except OSError as exc:
                    logger.warning(_("Cannot synchronize file %r: %s"),
                                   self._SESSION_FILE_NEXT, exc)
            finally:
                # Close the new session file
                logger.debug(_("Closing descriptor %d"), next_session_fd)
                os.close(next_session_fd)
            # Rename FILE_NEXT over FILE.
            #
            # Use renameat(2) to ensure that there is no race condition if the
            # location (directory) is being moved
            logger.debug(
                _("Renaming %r to %r"),
                self._SESSION_FILE_NEXT, self._SESSION_FILE)
            try:
                os.rename(self._SESSION_FILE_NEXT, self._SESSION_FILE,
                          src_dir_fd=location_fd, dst_dir_fd=location_fd)
            except Exception as exc:
                # Same as above, if we fail we need to unlink the next file
                # otherwise any other attempts will not be able to open() it
                # with O_EXCL flag.
                logger.warning(
                    _("Unable to rename/overwrite %r to %r: %r"),
                    self._SESSION_FILE_NEXT, self._SESSION_FILE, exc)
                # TRANSLATORS: unlinking as in deleting a file
                logger.warning(_("Unlinking %r"), self._SESSION_FILE_NEXT)
                os.unlink(self._SESSION_FILE_NEXT, dir_fd=location_fd)
            # Flush kernel buffers on the directory.
            #
            # This should ensure the rename operation is really on disk by now.
            # As noted above, this is essential for being able to survive
            # system crash immediately after exiting this method.

            # TRANSLATORS: please don't translate fsync()
            logger.debug(_("Calling fsync() on descriptor %d"), location_fd)
            try:
                os.fsync(location_fd)
            except OSError as exc:
                logger.warning(_("Cannot synchronize directory %r: %s"),
                               self.location, exc)
        finally:
            # Close the location directory
            logger.debug(_("Closing descriptor %d"), location_fd)
            os.close(location_fd)

    def break_lock(self):
        """
        Forcibly unlock the storage by removing a file created during
        atomic filesystem operations of save_checkpoint().

        This method might be useful if save_checkpoint()
        raises LockedStorageError. It removes the "next" file that is used
        for atomic rename.
        """
        _next_session_pathname = os.path.join(
            self.location, self._SESSION_FILE_NEXT)
        logger.debug(
            # TRANSLATORS: unlinking as in deleting a file
            # Please keep the 'next' string untranslated
            _("Forcibly unlinking 'next' file %r"), _next_session_pathname)
        os.unlink(_next_session_pathname)



# This file is part of Checkbox.
#
# Copyright 2013-2020 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
APIs for working with providers.

:mod:`plainbox.impl.providers` -- providers package
===================================================

Providers are a mechanism by which PlainBox can enumerate jobs and test plans.
Currently there are only v1 (as in version one) providers that basically have
to behave as CheckBox itself (mini CheckBox forks for example)

V1 providers
------------

The first (current) version of PlainBox providers has the following properties,
this is also described by :class:`IProvider1`::

    * there is a directory with '.txt' or '.txt.in' files with RFC822-encoded
      job definitions. The definitions need a particular set of keys to work.

    * there is a directory with additional executables (added to PATH)

    * there is a directory with an additional python3 libraries (added to
      PYTHONPATH)
"""

import logging
import os

from plainbox.impl.providers.embedded_providers import (
    EmbeddedProvider1PlugInCollection)

logger = logging.getLogger("plainbox.providers.__init__")


class ProviderNotFound(LookupError):

    """ Exception used to report that a provider cannot be located. """


def get_providers(*, only_secure: bool=False) -> 'List[Provider1]':
    """
    Find and load all providers that are available.

    :param only_secure:
        (keyword only) Return only providers that are deemed secure.
    :returns:
        A list of Provider1 objects, created in no particular order.

    This function can be used to get a list of all available providers. Most
    applications will just want the default, regular list of providers, without
    bothering to restrict themselves to the secure subset.

    Those are the providers that can run jobs as root using the
    ``plainbox-trusted-launcher-1`` mechanism. Depending on the *policykit*
    Policy, those might start without prompting the user for the password. If
    you want to load only them, use the `only_secure` option.
    """
    from plainbox.impl.providers import special
    if only_secure:
        from plainbox.impl.secure.providers.v1 import all_providers
    else:
        from plainbox.impl.providers.v1 import all_providers
    all_providers.load()
    std_providers= [
        special.get_manifest(),
        special.get_exporters(),
        special.get_categories(),
    ] + all_providers.get_all_plugin_objects()
    if only_secure:
        return std_providers

    def qualified_name(provider):
        return "{}:{}".format(provider.namespace, provider.name)
    sideload_path = os.path.expandvars(os.path.join(
        '/var', 'tmp', 'checkbox-providers'))
    embedded_providers = EmbeddedProvider1PlugInCollection(sideload_path)
    loaded_provs = embedded_providers.get_all_plugin_objects()
    for p in loaded_provs:
        logger.warning("Using sideloaded provider: %s from %s", p, p.base_dir)
    sl_qual_names = [qualified_name(p) for p in loaded_provs]
    for std_prov in std_providers:
        if qualified_name(std_prov) in sl_qual_names:
            # this provider got overriden by sideloading
            # so let's not load the original one
            continue
        loaded_provs.append(std_prov)
    if not loaded_provs:
        message = '\n'.join(( _("No providers found! Paths searched:"),
            *all_providers.provider_search_paths))
        raise SystemExit(message)
    return loaded_provs



# This file is part of Checkbox.
#
# Copyright 2016 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
Checkbox Launcher Interpreter Application
"""

import gettext
import logging
import os
import subprocess
import sys

from plainbox.impl.jobcache import ResourceJobCache
from plainbox.impl.launcher import DefaultLauncherDefinition
from plainbox.impl.launcher import LauncherDefinition
from plainbox.impl.session.assistant import SessionAssistant

from checkbox_ng.config import load_configs
from checkbox_ng.launcher.subcommands import (
    Launcher, List, Run, StartProvider, Submit, ListBootstrapped,
    TestPlanExport
)
from checkbox_ng.launcher.check_config import CheckConfig
from checkbox_ng.launcher.merge_reports import MergeReports
from checkbox_ng.launcher.merge_submissions import MergeSubmissions
from checkbox_ng.launcher.master import RemoteMaster
from checkbox_ng.launcher.slave import RemoteSlave


_ = gettext.gettext

_logger = logging.getLogger("checkbox-cli")


class Context:
    def __init__(self, args, sa):
        self.args = args
        self.sa = sa


def main():
    import argparse
    commands = {
        'check-config': CheckConfig,
        'launcher': Launcher,
        'list': List,
        'run': Run,
        'startprovider': StartProvider,
        'submit': Submit,
        'list-bootstrapped': ListBootstrapped,
        'merge-reports': MergeReports,
        'merge-submissions': MergeSubmissions,
        'tp-export': TestPlanExport,
        'service': RemoteSlave,
        'remote': RemoteMaster,
    }
    deprecated_commands = {
        'slave': 'service',
        'master': 'remote',
    }

    known_cmds = list(commands.keys())
    known_cmds += list(deprecated_commands.keys())
    known_cmds += ['-h', '--help']
    if not (set(known_cmds) & set(sys.argv[1:])):
        sys.argv.insert(1, 'launcher')

    for i, arg in enumerate(sys.argv):
        if arg in deprecated_commands:
            sys.argv[i] = deprecated_commands[arg]
            print()
            print('WARNING: "{}" deprecated'.format(arg), end='')
            print(' please use "{}" instead.'.format(
                deprecated_commands[arg]), end='\n\n')

    top_parser = argparse.ArgumentParser()
    top_parser.add_argument('-v', '--verbose', action='store_true', help=_(
        'print more logging from checkbox'))
    top_parser.add_argument('--debug', action='store_true', help=_(
        'print debug messages from checkbox'))
    top_parser.add_argument('--clear-cache', action='store_true', help=_(
        'remove cached results from the system'))
    top_parser.add_argument('--clear-old-sessions', action='store_true', help=_(
        "remove previous sessions' data"))
    top_parser.add_argument('--version', action='store_true', help=_(
        "show program's version information and exit"))
    top_parser.add_argument('subcommand', help=_("subcommand to run"),
                            choices=commands.keys())
    # parse all the cli invocation until a subcommand is found
    # subcommand doesn't start with a '-'
    subcmd_index = 1
    for i, arg in enumerate(sys.argv[1:]):
        if not arg.startswith('-'):
            subcmd_index = i + 1
            break
    args = top_parser.parse_args(sys.argv[1:subcmd_index + 1])
    subcmd_parser = argparse.ArgumentParser()
    subcmd = commands[args.subcommand]()
    subcmd.register_arguments(subcmd_parser)
    sub_args = subcmd_parser.parse_args(sys.argv[subcmd_index + 1:])
    sa = SessionAssistant(
        "com.canonical:checkbox-cli",
        "0.99",
        "0.99",
        ["restartable"],
    )
    ctx = Context(sub_args, sa)
    try:
        socket.getaddrinfo('localhost', 443)  # 443 for HTTPS
    except Exception:
        pass
    if '--clear-cache' in sys.argv:
        ResourceJobCache().clear()
    if '--clear-old-sessions' in sys.argv:
        old_sessions = [s[0] for s in sa.get_old_sessions()]
        sa.delete_sessions(old_sessions)
    if args.verbose:
        logging_level = logging.INFO
        logging.basicConfig(level=logging_level)
    if args.debug:
        logging_level = logging.DEBUG
        logging.basicConfig(level=logging_level)
    subcmd.invoked(ctx)



# This file is part of Checkbox.
#
# Copyright 2019 Canonical Ltd.
# Written by:
#   Jonathan Cave <jonathan.cave@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

import importlib.util
import os


def main():
    manage_f = os.path.join(os.getcwd(), 'manage.py')
    if not os.path.exists(manage_f):
        raise SystemExit('Could not find manage.py in current directory.'
                         ' Is this a plainbox provider?')
    spec = importlib.util.spec_from_file_location('setup', manage_f)
    foo = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(foo)



# This file is part of Checkbox.
#
# Copyright 2017-2020 Canonical Ltd.
# Written by:
#   Maciej Kisielewski <maciej.kisielewski@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.
"""
This module provides functions to handle sudo password handling.
The password stored by this module is not encrypted in any way and may be
transmitted over network in plaintext, so it's up to the operator to use
secure connection.
"""

import gc
import getpass
import hashlib
import logging
import os
import sys

from plainbox.i18n import gettext as _
from subprocess import check_call, CalledProcessError, DEVNULL, SubprocessError


logger = logging.getLogger("sudo_broker")


def is_passwordless_sudo():
    """
    Check if system can run sudo without pass.
    """
    if os.geteuid() == 0:
        # even though we run as root, we still may need to use sudo to switch
        # to a normal user for jobs not requiring root, so let's see if sudo
        # actually works.
        try:
            check_call(['sudo', '-A', '-k', 'true'], stdout=DEVNULL, stderr=DEVNULL)
        except (SubprocessError, OSError)  as exc:
            logger.error(_("Unable to run sudo %s"), exc)
            raise SystemExit(1)
        return True
    # running sudo with -A will try using ASKPASS envvar that should specify
    # the program to use when asking for password
    # If the system is configured to not ask for password, this will silently
    # succeed. If the pass is required, it'll return 1 and not ask for pass,
    # as the askpass program is not provided
    try:
        check_call(['sudo', '-A', '-k', 'true'], stdout=DEVNULL, stderr=DEVNULL)
    except CalledProcessError:
        return False
    return True


def validate_pass(password):
    cmd = ['sudo', '--prompt=', '--reset-timestamp', '--stdin',
           '--user', 'root', 'true']
    r, w = os.pipe()
    os.write(w, password + b"\n")
    os.close(w)
    try:
        check_call(cmd, stdin=r, stdout=DEVNULL, stderr=DEVNULL)
        return True
    except CalledProcessError:
        return False


class SudoPasswordProvider:
    def __init__(self):
        self._sudo_password = None
        self._already_checked = False
        self._is_passwordless = False

    @property
    def is_passwordless(self):
        if not self._already_checked:
            self._is_passwordless = is_passwordless_sudo()
            self._already_checked = True
        return self._is_passwordless

    def get_sudo_password(self):
        if self.is_passwordless:
            return None
        if self._sudo_password:
            return self._sudo_password
        pass_is_correct = False
        while not pass_is_correct:
            prompt = 'Enter sudo password:\n'
            password = getpass.getpass(prompt).encode(sys.stdin.encoding)
            pass_is_correct = validate_pass(password)
            if not pass_is_correct:
                print('Sorry, try again.')
        self._sudo_password = password
        return password

sudo_password_provider = SudoPasswordProvider()
